//############################################################################
//
// EZFB ~ Linux Frame Buffer API ~
//
// by James Lehman
// james@akrobiz.com
//
// began: Feb. 2001
//
// EZFB is intended to give C programmers an easy-to-use library of functions
// for drawing points, lines and strings of characters into the Linux frame
// buffer. It also provides a means of displaying bitmap images on any portion
// of the screen and capturing the contents of any portion of the screen to
// a standard bitmap graphics file. All of this is embodied in a set of
// functions that are polymorphic with regard to the characteristics of the
// frame buffer and the bitmap files used. In other words, it makes no
// difference what the screen height or width is or what the color depth is,
// all of the function calls are the same and will have appropriate and
// predictable effects.
//
// This software is distributed in the hope that it will be useful, but
// without any warranty; without even the implied warranty of merchantability
// or fitness for a particular purpose. See the GNU General Public License
// for more details.
//
//############################################################################
#include "ezfb_colors.h"
#include "ezfb_pixels.h"

//############################################################################
int ezfb_put_pixel__1bit(struct ezfb* fb, u_int  x, u_int  y, u_int  c)
{
    u_int xx;
EZFB_NIT_PICKING_FUNCTION_CALL_1
    xx = x / 8;
    if(    (&(fb->screen_row_uchar[y][xx]) >= fb->mem_uchar    )
        && (&(fb->screen_row_uchar[y][xx]) <= fb->mem_last_uchar)
      )
    switch(x % 8)
    {
        case 0: fb->screen_row_uchar[y][xx] = (fb->screen_row_uchar[y][xx] & ~0x01) | ((c & 0x01)     ); break;
        case 1: fb->screen_row_uchar[y][xx] = (fb->screen_row_uchar[y][xx] & ~0x02) | ((c & 0x01) << 1); break;
        case 2: fb->screen_row_uchar[y][xx] = (fb->screen_row_uchar[y][xx] & ~0x04) | ((c & 0x01) << 2); break;
        case 3: fb->screen_row_uchar[y][xx] = (fb->screen_row_uchar[y][xx] & ~0x08) | ((c & 0x01) << 3); break;
        case 4: fb->screen_row_uchar[y][xx] = (fb->screen_row_uchar[y][xx] & ~0x10) | ((c & 0x01) << 4); break;
        case 5: fb->screen_row_uchar[y][xx] = (fb->screen_row_uchar[y][xx] & ~0x20) | ((c & 0x01) << 5); break;
        case 6: fb->screen_row_uchar[y][xx] = (fb->screen_row_uchar[y][xx] & ~0x40) | ((c & 0x01) << 6); break;
        case 7: fb->screen_row_uchar[y][xx] = (fb->screen_row_uchar[y][xx] & ~0x80) | ((c & 0x01) << 7); break;
    }
EZFB_NIT_PICKING_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_put_pixel__2bit(struct ezfb* fb, u_int  x, u_int  y, u_int  c)
{
    u_int xx;
EZFB_NIT_PICKING_FUNCTION_CALL_1
    xx = x / 4;
    if(    (&(fb->screen_row_uchar[y][xx]) >= fb->mem_uchar    )
        && (&(fb->screen_row_uchar[y][xx]) <= fb->mem_last_uchar)
      )
    switch(x % 4)
    {
        case 0: fb->screen_row_uchar[y][xx] = (fb->screen_row_uchar[y][xx] & ~0x03) | ((c & 0x03)     ); break;
        case 1: fb->screen_row_uchar[y][xx] = (fb->screen_row_uchar[y][xx] & ~0x0c) | ((c & 0x03) << 2); break;
        case 2: fb->screen_row_uchar[y][xx] = (fb->screen_row_uchar[y][xx] & ~0x30) | ((c & 0x03) << 4); break;
        case 3: fb->screen_row_uchar[y][xx] = (fb->screen_row_uchar[y][xx] & ~0xc0) | ((c & 0x03) << 6); break;
    }
EZFB_NIT_PICKING_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_put_pixel__4bit(struct ezfb* fb, u_int  x, u_int  y, u_int  c)
{
    u_int xx;
EZFB_NIT_PICKING_FUNCTION_CALL_1
    xx = x / 2;
    if(    (&(fb->screen_row_uchar[y][xx]) >= fb->mem_uchar    )
        && (&(fb->screen_row_uchar[y][xx]) <= fb->mem_last_uchar)
      )
    fb->screen_row_uchar[y][xx] =    (x & 0x01) // odd
                                   ? ((fb->screen_row_uchar[y][xx] & 0x0F) | ((c & 0x0F) << 4))
                                   : ((fb->screen_row_uchar[y][xx] & 0xF0) |  (c & 0x0F)      );
EZFB_NIT_PICKING_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_put_pixel__8bit(struct ezfb* fb, u_int  x, u_int  y, u_int  c)
{
EZFB_NIT_PICKING_FUNCTION_CALL_1
    if(    (&(fb->screen_row_uchar[y][x]) >= fb->mem_uchar    )
        && (&(fb->screen_row_uchar[y][x]) <= fb->mem_last_uchar)
      )
    fb->screen_row_uchar[y][x] = c;
EZFB_NIT_PICKING_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_put_pixel_16bit(struct ezfb* fb, u_int  x, u_int  y, u_int  c)
{
EZFB_NIT_PICKING_FUNCTION_CALL_1
    if(    (&(fb->screen_row_ushort[y][x]) >= (u_short*)(fb->mem_uchar    ))
        && (&(fb->screen_row_ushort[y][x]) <= (u_short*)(fb->mem_last_uchar))
      )
        fb->screen_row_ushort[y][x] = c;
EZFB_NIT_PICKING_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_put_pixel_24bit(struct ezfb* fb, u_int  x, u_int  y, u_int  c)
{
    u_int xx;
EZFB_NIT_PICKING_FUNCTION_CALL_1
    xx = x * 3;
    if(    (&(fb->screen_row_uchar[y][xx]) >= fb->mem_uchar    )
        && (&(fb->screen_row_uchar[y][xx]) <= fb->mem_last_uchar)
      )
    {
        fb->screen_row_uchar[y][xx + 2] = ezfb_get_r_from_rgb(fb, c);
        fb->screen_row_uchar[y][xx + 1] = ezfb_get_g_from_rgb(fb, c);
        fb->screen_row_uchar[y][xx    ] = ezfb_get_b_from_rgb(fb, c);
    }
EZFB_NIT_PICKING_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_put_pixel_32bit(struct ezfb* fb, u_int  x, u_int  y, u_int  c)
{
EZFB_NIT_PICKING_FUNCTION_CALL_1
    if(    (&(fb->screen_row_uint[y][x]) >= (u_int*)(fb->mem_uchar    ))
        && (&(fb->screen_row_uint[y][x]) <= (u_int*)(fb->mem_last_uchar))
      )
        fb->screen_row_uint[y][x] = c;
EZFB_NIT_PICKING_FUNCTION_RETURN(ret)
}

//############################################################################
u_int ezfb_get_pixel__1bit(struct ezfb* fb, u_int  x, u_int  y)
{
    u_int xx;
    if(fb->exists && (x < fb->Var.xres_virtual) && (y < fb->Var.yres_virtual))
    {
        xx = x / 8;
        switch(x % 8)
        {
            case 0: return (fb->screen_row_uchar[y][xx] & 0x01)     ;
            case 1: return (fb->screen_row_uchar[y][xx] & 0x02) >> 1;
            case 2: return (fb->screen_row_uchar[y][xx] & 0x04) >> 2;
            case 3: return (fb->screen_row_uchar[y][xx] & 0x08) >> 3;
            case 4: return (fb->screen_row_uchar[y][xx] & 0x10) >> 4;
            case 5: return (fb->screen_row_uchar[y][xx] & 0x20) >> 5;
            case 6: return (fb->screen_row_uchar[y][xx] & 0x40) >> 6;
            case 7: return (fb->screen_row_uchar[y][xx] & 0x80) >> 7;
        }
    }
    return 0;
}

//############################################################################
u_int ezfb_get_pixel__2bit(struct ezfb* fb, u_int  x, u_int  y)
{
    u_int xx;
    if(fb->exists && (x < fb->Var.xres_virtual) && (y < fb->Var.yres_virtual))
    {
        xx = x / 4;
        switch(x % 4)
        {
            case 0: return (fb->screen_row_uchar[y][xx] & 0x03);
            case 1: return (fb->screen_row_uchar[y][xx] & 0x0c) >> 2;
            case 2: return (fb->screen_row_uchar[y][xx] & 0x30) >> 4;
            case 3: return (fb->screen_row_uchar[y][xx] & 0xc0) >> 6;
        }
    }
    return 0;
}

//############################################################################
u_int ezfb_get_pixel__4bit(struct ezfb* fb, u_int  x, u_int  y)
{
    u_int xx;
    if(fb->exists && (x < fb->Var.xres_virtual) && (y < fb->Var.yres_virtual))
    {
        xx = x / 2;
        return     (x & 0x01) // odd
                 ? ( fb->screen_row_uchar[y][xx] & 0x0F)
                 : ((fb->screen_row_uchar[y][xx] & 0xF0) >> 4);
    }
    return 0;
}

//############################################################################
u_int ezfb_get_pixel__8bit(struct ezfb* fb, u_int  x, u_int  y)
{
    if(fb->exists && (x < fb->Var.xres_virtual) && (y < fb->Var.yres_virtual))
        return (u_int)fb->screen_row_uchar[y][x];
    return 0;
}

//############################################################################
u_int ezfb_get_pixel_16bit(struct ezfb* fb, u_int  x, u_int  y)
{
    if(fb->exists && (x < fb->Var.xres_virtual) && (y < fb->Var.yres_virtual))
        return (u_int)fb->screen_row_ushort[y][x];
    return 0;
}

//############################################################################
u_int ezfb_get_pixel_24bit(struct ezfb* fb, u_int  x, u_int  y)
{
    u_int xx;
    if(fb->exists && (x < fb->Var.xres_virtual) && (y < fb->Var.yres_virtual))
    {
        xx = x * 3;
        return    (fb->screen_row_uchar[y][xx    ] << 16)
                | (fb->screen_row_uchar[y][xx + 1] <<  8)
                | (fb->screen_row_uchar[y][xx + 2]      );
    }
    return 0;
}

//############################################################################
u_int ezfb_get_pixel_32bit(struct ezfb* fb, u_int  x, u_int  y)
{
    if(fb->exists && (x < fb->Var.xres_virtual) && (y < fb->Var.yres_virtual))
        return fb->screen_row_uint[y][x];
    return 0;
}

//############################################################################
int ezfb_put_line(struct ezfb* fb, int x1, int y1, int x2, int y2, int c)
{
    int dx  = x2 - x1;
    int dy  = y2 - y1;
    int n, b;
    float m;
EZFB_NIT_PICKING_FUNCTION_CALL_1
    if(0 == dx)
    {
        if(dy < 0)
            for(n = y1; n >= y2; n--)
                ret &= fb->put_pixel(fb, x1, n, c);
        else
            for(n = y1; n <= y2; n++)
                ret &= fb->put_pixel(fb, x1, n, c);
    }
    else if(0 == dy)
    {
        if(dx < 0)
            for(n = x1; n >= x2; n--)
                ret &= fb->put_pixel(fb, n, y1, c);
        else
            for(n = x1; n <= x2; n++)
                ret &= fb->put_pixel(fb, n, y1, c);
    }
    else if(abs(dx) >= abs(dy))
    {
        m = ((float)dy) / ((float)dx);
        b = (int)(y1 - m * x1);
        if(dx < 0)
            for(n = x1; n >= x2; n--)
                ret &= fb->put_pixel(fb, n, (u_int)floor(n * m + b + 0.5), c);
        else
            for(n = x1; n <= x2; n++)
                ret &= fb->put_pixel(fb, n, (u_int)floor(n * m + b + 0.5), c);
    }
    else
    {
        m = ((float)dx) / ((float)dy);
        b = (int)(x1 - m * y1);
        if(dy < 0)
            for(n = y1; n >= y2; n--)
                ret &= fb->put_pixel(fb, (u_int)floor(n * m + b + 0.5), n, c);
        else
            for(n = y1; n <= y2; n++)
                ret &= fb->put_pixel(fb, (u_int)floor(n * m + b + 0.5), n, c);
    }
EZFB_NIT_PICKING_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_put_ray(struct ezfb* fb, int x, int y, float r, float t, int c)
{
    return ezfb_put_line(   fb
                          , x
                          , y
                          , x + floor((cos(t) * r) + 0.5)
                          , y + floor((sin(t) * r) + 0.5)
                          , c
                        );
}

//############################################################################
int ezfb_put_rectangle(struct ezfb* fb, int x1, int y1, int x2, int y2, int c, int filled)
{
EZFB_NIT_PICKING_FUNCTION_CALL_1
    if(filled)
    {
        int y, step;
        step = (y1 <= y2) ? 1 : -1;
        if((x1 != x2) && (y1 != y2))
            for(y = y1; y <= y2; y += step)
                ret &= ezfb_put_line(fb, x1, y, x2, y, c);
        else if(x1 != x2)
            ret &= ezfb_put_line(fb, x1, y1, x2, y1, c);
        else
            ret &= fb->put_pixel(fb, x1, y1, c);
    }
    else
    {
        if((x1 != x2) && (y1 != y2))
        {
            ret &= ezfb_put_line(fb, x1, y1, x2, y1, c);
            ret &= ezfb_put_line(fb, x1, y2, x2, y2, c);
            ret &= ezfb_put_line(fb, x1, y1, x1, y2, c);
            ret &= ezfb_put_line(fb, x2, y1, x2, y2, c);
        }
        else if(x1 != x2)
            ret &= ezfb_put_line(fb, x1, y1, x2, y1, c);
        else
            ret &= fb->put_pixel(fb, x1, y1, c);
    }
EZFB_NIT_PICKING_FUNCTION_RETURN(ret)
}
//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################

